# -*- Mode: Python -*-

import getopt
import re, sys
import asyncore
import os
from medusa import http_server
from medusa import default_handler
from medusa import logger
from medusa import script_handler
from medusa import filesys
import medusa.producers
import medusa.http_date

import random
import imp
import posix
import time
import stat
import urllib
import traceback


"""
TODO: 
    * contexts
    * session clearup
    * xml file parsing
    * apache proxy testing
"""

use_cookies = 1

GLOBAL_TEMP_DIR="/tmp/"

verbose = 0

class WebContext:
    def __init__(self, name):
        self.name = name
        self.files = []
        self.modules = {}
    def addFile(self, file):
        self.files += [file]
    def setRoot(self, root):
        if root[-1] != os.sep:
            root += os.sep
        self.root = root

class WebFile:
    def __init__(self, context, filename):
        self.context = context
        if filename[0] == os.sep:
            filename = filename[1:]
        self.filename = filename
        self.handlers = []
    def addHandler(self, handler):
        self.handlers += [handler]
    def getFileName(self):
        return self.context.root + self.filename

class WebHandler:
    def __init__(self, file, function):
        self.file = file
        self.function = function
        self.patterns = []
    def addPattern(self, pattern):
        self.patterns += [pattern]

class WebPattern:
    def __init__(self, handler, pattern):
        self.handler = handler 
        self.pattern = pattern 
        if not pattern.endswith('$'):
            pattern = pattern + "$"
        self.compiled = re.compile(pattern)
    def getPattern(self):
        return self.compiled
    def getPatternString(self):
        return self.pattern

def qualify_path(path):
    if path[-1] != os.sep:
        path += os.sep;
    return path

def read_ini_file(filename):
    lineno = 0
    fi = open(filename, "rb")
    file = None
    function = None
    context = None
    contexts = []
    maindir = "" + os.sep
    for line in fi.readlines():
        lineno=lineno+1
        # remove comments
        hashpos = line.find("#")
        if hashpos>=0:
            line = line[0:hashpos]
        # remove whitespace
        line = line.strip()
        
        if line == "":
            continue #skip empty line

        equals = line.find(":")
        if equals<0:
            continue
        key = line[0:equals].strip()
        value = line[equals+1:].strip()
        if key == "base":
            maindir = qualify_path(value)
        elif key == "context":
            contextname = value
            context = WebContext(contextname)
            contexts += [context]
        elif key == "root":
            #if not value.startswith(os.sep) and not value[1]==':':
            #    value = maindir + value
            if value.startswith(os.sep):
                value = value[1:]
            value = maindir + value
            context.setRoot(value)
        elif key == "file":
            filename = value
            file = WebFile(context, filename)
            context.addFile(file)
        elif key == "handler":
            function = value
            handler = WebHandler(file, function)
            file.addHandler(handler)
        elif key == "pattern":
            patternstr = value
            pattern = WebPattern(handler, patternstr)
            handler.addPattern(pattern)
        else:
            raise "Syntax error in line "+str(lineno)+" of file "+filename+":\n"+line
    fi.close()
    return contexts

def create_session_id():
    pid = abs((str(posix.getpid())).__hash__())
    now = abs((str(time.time())).__hash__())
    rand = abs((str(random.random())).__hash__())
    x = "abcdefghijklmnopqrstuvwxyz0123456789"
    result = ""
    for a in range(0,6):
        result += x[pid%36]
        pid = pid / 36
    result += "-"
    for a in range(0,6):
        result += x[now%36]
        now = now / 36
    result += "-"
    for a in range(0,6):
        result += x[rand%36]
        rand = rand / 36
    return result

def g(h,k,d=None):
    if k in h:
        return h[k]
    else:
        return d

def split_headers(string):
    return string.split("\r\n")

def headers_to_map(mylist):
    headers={}
    for h in mylist:
        try:
            i = h.index(':')
        except:
            i = -1
        if i >= 0:
            key = h[0:i].lower()
            value = h[i+1:]
            if len(value)>0 and value[0] == ' ':
                value = value[1:]
            headers[key] = value
        else:
            if len(h.strip())>0:
                print "invalid header:",h
    return headers
                    
def parse_semicolon_parameters(params):
    params = params.split("; ")
    parmap = {}
    for a in params:
        if '=' in a:
            key,value = a.split('=')
            if value.startswith('"') and value.endswith('"'):
                value = value[1:-1]
            parmap[key] = value
    return parmap

#def check_outdated_modules():
#    for m in sys.modules.keys():
#        if m not in original:
#            # unless they are part of standard library
#            mod = sys.modules[m]
#            if hasattr(mod, "__file__"):
#                path, x = os.path.split(mod.__file__)
#                if path != stdlib:
#                    del sys.modules[m]

HTTP_CONTINUE                     = 100
HTTP_SWITCHING_PROTOCOLS          = 101
HTTP_PROCESSING                   = 102
HTTP_OK                           = 200
HTTP_CREATED                      = 201
HTTP_ACCEPTED                     = 202
HTTP_NON_AUTHORITATIVE            = 203
HTTP_NO_CONTENT                   = 204
HTTP_RESET_CONTENT                = 205
HTTP_PARTIAL_CONTENT              = 206
HTTP_MULTI_STATUS                 = 207
HTTP_MULTIPLE_CHOICES             = 300
HTTP_MOVED_PERMANENTLY            = 301
HTTP_MOVED_TEMPORARILY            = 302
HTTP_SEE_OTHER                    = 303
HTTP_NOT_MODIFIED                 = 304
HTTP_USE_PROXY                    = 305
HTTP_TEMPORARY_REDIRECT           = 307
HTTP_BAD_REQUEST                  = 400
HTTP_UNAUTHORIZED                 = 401
HTTP_PAYMENT_REQUIRED             = 402
HTTP_FORBIDDEN                    = 403
HTTP_NOT_FOUND                    = 404
HTTP_METHOD_NOT_ALLOWED           = 405
HTTP_NOT_ACCEPTABLE               = 406
HTTP_PROXY_AUTHENTICATION_REQUIRED= 407
HTTP_REQUEST_TIME_OUT             = 408
HTTP_CONFLICT                     = 409
HTTP_GONE                         = 410
HTTP_LENGTH_REQUIRED              = 411
HTTP_PRECONDITION_FAILED          = 412
HTTP_REQUEST_ENTITY_TOO_LARGE     = 413
HTTP_REQUEST_URI_TOO_LARGE        = 414
HTTP_UNSUPPORTED_MEDIA_TYPE       = 415
HTTP_RANGE_NOT_SATISFIABLE        = 416
HTTP_EXPECTATION_FAILED           = 417
HTTP_UNPROCESSABLE_ENTITY         = 422
HTTP_LOCKED                       = 423
HTTP_FAILED_DEPENDENCY            = 424
HTTP_INTERNAL_SERVER_ERROR        = 500
HTTP_NOT_IMPLEMENTED              = 501
HTTP_BAD_GATEWAY                  = 502
HTTP_SERVICE_UNAVAILABLE          = 503
HTTP_GATEWAY_TIME_OUT             = 504
HTTP_VERSION_NOT_SUPPORTED        = 505
HTTP_VARIANT_ALSO_VARIES          = 506
HTTP_INSUFFICIENT_STORAGE         = 507
HTTP_NOT_EXTENDED                 = 510

class MyFile:
    def __init__(self,fieldname, parammap,filename,content_type):
        #print "field-name:", fieldname
        #print "filename:", filename
        #print "content-type:", content_type
        self.fieldname = fieldname
        self.parammap = parammap
        self.filename = filename
        self.content_type = content_type
        self.tempname = GLOBAL_TEMP_DIR+str(random.random())
        self.filesize = 0
        self.fi = open(self.tempname, "wb")
    def adddata(self,data):
        self.filesize += len(data)
        #print len(data), "bytes of file data"
        self.fi.write(data)
        #print "***"+data+"***"
    def close(self):
        self.fi.close()
        self.parammap[self.fieldname] = self
        del self.fieldname
        del self.parammap
        del self.fi
    def __str__(self):
        return "file %s (%s), %d bytes, content-type: %s" % (self.filename, self.tempname, self.filesize, self.content_type)

class MyField:
    def __init__(self,fieldname,parammap):
        self.fieldname = fieldname
        self.data = ""
        self.parammap = parammap
    def adddata(self,data):
        self.data += data
    def close(self):
        self.parammap[self.fieldname] = self.data
        del self.data
        del self.parammap


class simple_input_collector:
    def __init__ (self, handler, request, length):
        self.request = request
        self.length = length
        self.handler = handler
        request.channel.set_terminator(length)
        self.data = ""

    def collect_incoming_data (self, data):
        self.data += data

    def found_terminator(self):
        self.request.channel.set_terminator('\r\n\r\n')
        self.request.collector = None
        d=self.data;del self.data
        r=self.request;del self.request
        parameters={}
        data = d.split('&')
        for e in data:
            if '=' in e:
                key,value = e.split('=')
                parameters[key] = urllib.unquote_plus(value)
            else:
                if len(e.strip())>0:
                    print "Unknown parameter: ",e
        self.handler.continue_request(r,parameters)

class upload_input_collector:
    def __init__ (self, handler, request, length, boundary):
        self.request = request
        self.length = length
        self.handler = handler
        self.boundary = boundary
        request.channel.set_terminator(length)
        self.data = ""
        self.pos = 0
        self.start_marker = "--"+boundary+"\r\n"
        self.end_marker = "--"+boundary+"--\r\n"
        self.prefix = "--"+boundary
        self.marker = "\r\n--"+boundary
        self.header_end_marker = "\r\n\r\n"
        self.current_file = None
        self.boundary = boundary
        #print "START parsing",self.boundary
        self.file = None
        self.parameters = {}
        self.files = []

    def startFile(self,headers):
        fieldname = None
        filename = None
        if self.file is not None:
            raise "Illegal state"
        if "content-disposition" in headers:
            cd = headers["content-disposition"]
            l = parse_semicolon_parameters(cd)
            if "name" in l:
                fieldname = l["name"]
            if "filename" in l:
                filename = l["filename"]
        if "content-type" in headers:
            content_type = headers["content-type"]
            self.file = MyFile(fieldname,self.parameters,filename,content_type)
            self.files += [self.file]
        else:
            self.file = MyField(fieldname,self.parameters)

    def collect_incoming_data (self, newdata):
        self.pos += len(newdata)
        self.data += newdata
       
        while len(self.data)>0:
            if self.data.startswith(self.end_marker):
                self.data = self.data[len(self.end_marker):]
                #print "DONE with",self.boundary
                if self.file is not None:
                    self.file.close()
                    self.file = None
                return
            elif self.data.startswith(self.start_marker):
                try:
                    i = self.data.index(self.header_end_marker, len(self.start_marker))
                except:
                    i = -1
                if i>=0:
                    headerstr = self.data[len(self.start_marker):i+2]
                    headers = headers_to_map(split_headers(headerstr))
                    self.startFile(headers)
                    self.data = self.data[i+len(self.header_end_marker):]
                else:
                    return # wait for more data (inside headers)
            elif self.data.startswith(self.prefix):
                #rare case- marker is there, end/start suffix still outstanding- wait for more data,
                #so we know whether this is a start or end marker
                return
            else:
                try:
                    # TODO: python's str.index() function could use some speedup
                    bindex = self.data.index(self.marker)
                    self.file.adddata(self.data[0:bindex])
                    self.file.close()
                    self.file = None
                    self.data = self.data[bindex+2:] # cut to position after \r\n
                except ValueError: #not found
                    if(len(self.data) <= len(self.marker)):
                        return #wait for more data before we make a decision or pass through data
                    else:
                        self.file.adddata(self.data[0:-len(self.marker)])
                        self.data = self.data[-len(self.marker):]

    def found_terminator(self):
        if len(self.data)>0:# and self.file is not None:
            if self.file is not None:
                self.file.close()
                self.file = None
            raise "Unfinished/malformed multipart request"
        #if len(self.data)>0:
        #    self.collect_incoming_data("") #TODO: shouldn't be needed anymore
        if self.file is not None:
            self.file.close()
            self.file = None
            
        self.request.collector = None
        self.request.channel.set_terminator('\r\n\r\n')
        d=self.data;del self.data
        r=self.request;del self.request
        self.handler.continue_request(r,self.parameters)
        for f in self.files:
            os.unlink(f.tempname)

class MyRequest:
    def __init__(self,context,request,sessionid):
        self.request = request
        self.sessionid = sessionid
    def write(self,text):
        self.request.push(text)
    def setStatus(self,status):
        self.request.reply_code = status
    def makeLink(self,page,params=None):
        # TODO: handle context
        query = ""
        if params is not None:
            first = 1
            for k,v in params.items():
                if first:
                    query += "?"
                else:
                    query += "&"
                query += default_handler.quote(k)+"="+default_handler.quote(v)
                first = 0
        return page+";"+self.sessionid+query

    def sendFile(self,path,content_type):

        file_length = os.stat(path)[stat.ST_SIZE]
        ims = default_handler.get_header_match (default_handler.IF_MODIFIED_SINCE, self.request.header)
        length_match = 1
        if ims:
            length = ims.group (4)
            if length:
                try:
                    length = string.atoi (length)
                    if length != file_length:
                        length_match = 0
                except:
                    pass
        ims_date = 0
        if ims:
            ims_date = medusa.http_date.parse_http_date (ims.group (1))

        try:
            mtime = os.stat (path)[stat.ST_MTIME]
        except:
            self.request.error (404)
            return
        if length_match and ims_date:
            if mtime <= ims_date:
                self.request.reply_code = 304
                return
        try:
            file = open (path, 'rb')
        except IOError:
            self.request.error (404)
            return

        self.request['Last-Modified'] = medusa.http_date.build_http_date (mtime)
        self.request['Content-Length'] = file_length
        self.request['Content-Type'] = content_type
        self.request['Connection'] = 'close';
        if self.request.command == 'GET':
            self.request.push (medusa.producers.file_producer(file))
        return

    def setCookie(self, name, value, expire=None):
        if 'Set-Cookie' in self.request.reply_headers:
            # TODO: fix this, possibly by forking medusa
            raise "Can only set one cookie per request"
        if expire is None:
            self.request['Set-Cookie'] = name+'='+value;
        else:
            datestr = time.strftime("%a, %d-%b-%Y %H:%M:%S GMT", time.gmtime(expire))
            self.request['Set-Cookie'] = name+'='+value+'; expires='+datestr; #+'; path=PATH; domain=DOMAIN_NAME; secure';


class Session:
    # TODO: clear sessions after some time
    def __init__(self, id):
        self.id = id
        self.data = {}
    def use(self):
        self.lastuse = time.time()
    def __setitem__ (self, key, value):
        self.data[key] = value
    def __getitem__ (self, key):
        return self.data[key]
    def __hasitem__ (self, key):
        try: 
            self.data[key]
        except:
            return 0
        return 1
    def __delitem__(self, key):
        del self.data[key]



BASENAME = re.compile("([^/]*/)*([^/.]*)(.py)?")
MULTIPART = re.compile ('multipart/form-data.*boundary=([^ ]*)', re.IGNORECASE)
SESSION_PATTERN = re.compile("^;[a-z0-9]{6}-[a-z0-9]{6}-[a-z0-9]{6}$")

class Myhandler:
    def __init__(self, contexts):
        self.contexts = contexts
        self.sessions = {}

    def match(self, request):
        path, params, query, fragment = request.split_uri()
        if verbose:
            print "===== request:",path,"====="
        return 1
    
    def handle_request (self, request):
        headers = headers_to_map(request.header)
        request.request_headers = headers

        size=g(headers,"content-length",None)

        if size and size != '0':
            size=int(size)
            ctype=g(headers,"content-type",None)
            b = MULTIPART.match(ctype)
            if b is not None:
                #print "Handling multipart request"
                request.type = "MULTIPART"
                boundary = b.group(1)
                request.collector = upload_input_collector(self,request,size,boundary)
            else:
                request.type = "POST"
                #print "Handling post request"
                request.collector = simple_input_collector(self,request,size)
        else:
            request.type = "GET"
            #print "Handling get request"
            self.continue_request(request, {})

    def continue_request(self, request, parameters):

        path, params, query, fragment = request.split_uri()
        #print path, params, query, fragment,data
        #print data

        if query is not None:
            if query[0] == '?':
                query=query[1:]
            query = query.split('&')
            for e in query:
                key,value = e.split('=')
                parameters[key] = urllib.unquote_plus(value) #_plus?
                print key,value

        cookies = {}
        if "cookie" in request.request_headers:
            cookiestr = request.request_headers["cookie"]
            items = cookiestr.split(';')
            for a in items:
                key,value = a.strip().split('=')
                cookies[key] = value

        request.Cookies = cookies

        sessionid = None
        if params is not None and SESSION_PATTERN.match(params):
            #existing session
            sessionid = params
            if sessionid[0] == ';':
                sessionid = sessionid[1:]
        elif use_cookies and "PSESSION" in cookies:
            sessionid = cookies["PSESSION"]
            #print "sessionid:", cookies["PSESSION"]

        if sessionid is not None:
            if sessionid in self.sessions:
                session = self.sessions[sessionid]
                session.use()
            else:
                #TODO: should we output an "invalid session" page instead?
                session = Session(sessionid)
                self.sessions[sessionid] = session
            #print "Session: "+session.id
        else:
            #no session yet
            sessionid = create_session_id()
            session = Session(sessionid)
            self.sessions[sessionid] = session


        request['Connection'] = 'close';

        maxlen = -1
        context = None
        for c in self.contexts:
            if path.startswith(c.name) and len(c.name)>maxlen:
                context = c
                maxlen = len(context.name)
        if context is None:
            request.error (404)
            return

        print "Request ",'"'+path+'"',"maps to context",context.name
        path = path[len(context.name):]
        if len(path)==0 or path[0] != '/':
            path = "/" + path

        req = MyRequest(context,request,sessionid)
        req.session = session
        req.path = path
        req.paramstring = params
        req.query = query
        req.fragment = fragment
        req.params = parameters
        req.type = request.type

        if use_cookies:
            req.setCookie('PSESSION', sessionid, time.time()+3600*2)

        request.channel.current_request = None

        #look for a handler
        function = None
        for pattern,call in context.pattern_to_function.items():
            if pattern.match(path):
                function,desc = call
                if verbose:
                    print "Request %s matches (%s)" % (req.path, desc)
        if function is None:
            for id,call in context.id_to_function.items():
                if req.path == id:
                    function,desc = call
                    if verbose:
                        print "Request %s matches handler (%s)" % (req.path, desc)
       
        if function is not None:
            #found a handler
            status = function(req)
            if status is not None and type(1)==type(status) and status>10:
                request.reply_code = status
            request.done()
            return
        else:
            #no handler- redirect to filesystem
            if verbose:
                print "Request %s matches no pattern" % req.path
            return context.dh.handle_request(request)

        
def start_server(INIFILE, WEBLOG, TEMPDIR):

    GLOBAL_TEMP_DIR = TEMPDIR
    
    contexts = read_ini_file(INIFILE)

    for context in contexts:
        context.dh = default_handler.default_handler (filesys.os_filesystem (context.root))
        context.pattern_to_function = {}
        context.id_to_function = {}
        for file in context.files:
            filename = file.getFileName()
            b = BASENAME.match(filename)
            if b is None:
                raise "Internal error with filename "+filename
            module = b.group(2)
            if module is None:
                raise "Internal error with filename "+filename

            if module in context.modules:
                m = context.modules[module]
            else:
                try:
                    #m = modules[module] = imp.load_source(module, filename)
                    (mfile,pathname,description) = imp.find_module(module, [context.root])
                    m = context.modules[module] = imp.load_module(module,mfile,pathname,description)
                except IOError:
                    raise "IOError while loading file "+filename
            for handler in file.handlers:
                try:
                    f = eval("m."+handler.function)
                    if f is None: 
                        raise
                except:
                    print "Unexpected error:", sys.exc_info()[0], sys.exc_info()[1]
                    traceback.print_tb(sys.exc_info()[2])
                    raise "No such function "+handler.function+" in file "+filename

                for pattern in handler.patterns:
                    patternstr = pattern.getPatternString()
                    desc = "pattern %s, file %s, function %s" % (patternstr,filename,handler.function)
                    desc2 = "file %s, function %s" % (filename,handler.function)
                    context.pattern_to_function[pattern.getPattern()] = (f,desc)
                    context.id_to_function["/"+handler.function] = (f,desc2)

    lg = None
    if WEBLOG is not None:
        fi = open(WEBLOG, "wb")
        lg = logger.file_logger (fi)

    ph = Myhandler(contexts)
    hs = http_server.http_server ('', 8081, logger_object = lg)

    hs.install_handler (ph)

    asyncore.loop()

if __name__ == '__main__':
    #start_server(INIFILE="web.txt", WEBLOG="weblog.txt", TEMPDIR="/tmp/")

    def usage():
        print
        print sys.argv[0] + " [initfile]"
        print
        print "-v, --verbose          Be more verbose"
        print "-d, --debug            Turn on debugging"
        print "-i, --init-file <path> Set the init file to use"
        print "-r, --root-directory   Set the root directory (web base) (default: ./)"
        print "-l, --log-file         Set the logging file to use"
        print "-t, --temp-path        Set the temporary directory (default: /tmp/)"
        print
        sys.exit(2)

    try:
        opts, args = getopt.getopt(sys.argv[1:], "hvdi:r:l:t:", ["help", "verbose", "debug", "init-file=", "root-directory=", "log-file=", "temp-path="])
    except getopt.GetoptError:
        usage()

    #print opts
    #print args

    verbose = 1
    init_file="web.cfg"
    log_file=None
    temp_path="/tmp/"
    root_directory="."

    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        if o in ("-v","--verbose"):
            verbose = 2
        if o in ("-q","--quiet"):
            verbose = 0
        if o in ("-d","--debug"):
            verbose = 3
        if o in ("-i","--init-file"):
            init_file = a
        if o in ("-l","--log-file"):
            log_file = a
        if o in ("-t","--temp-path"):
            temp_path = a


    print "-"*72
    print "Starting Athana..."
    print "Init-File:",init_file
    print "Log-File:",log_file
    print "Temp-Path:",temp_path
    print "-"*72

    start_server(INIFILE=init_file, WEBLOG=log_file, TEMPDIR=temp_path)

